#include "sox.h"
#include <stdexcept>
#include <fstream>
#include <sstream>
#include <iostream>
#include <vector>
#include <functional>
#include <numeric>
#include <cstring>

namespace {
// ensures that files will be closed in case of error
struct file_handle {
  sox_format_t* h_;

  file_handle(sox_format_t* h) : h_(h) {}
  ~file_handle() {
    if (h_)
      sox_close(h_);
  }
};

// ensures deletion of effects in case of error
struct effect_handle {
  sox_effect_t* h_;

  effect_handle(sox_effect_t* h) : h_(h) {}
  ~effect_handle() {
    if (h_)
      sox_delete_effect(h_);
  }
};

// ensures deletion of effect chain, and factor common logic when adding effects
struct chain_handle {
  // the effects chain
  sox_effects_chain_t* h_;

  // describes the next effect's input, it will be modified by add_effect
  sox_signalinfo_t sig_in_;

  // describes the chain's output
  sox_signalinfo_t* sig_out_;

  // true until the first effect has been added
  bool empty_;

  chain_handle(sox_encodinginfo_t* in, sox_encodinginfo_t* out,
               sox_signalinfo_t* sig_in, sox_signalinfo_t* sig_out)
    : h_(sox_create_effects_chain(in, out)), sig_in_(*sig_in),
      sig_out_(sig_out), empty_(true) {
    if (!h_)
      throw std::runtime_error("Error creating chain");
  }

  ~chain_handle() {
    if (h_)
      sox_delete_effects_chain(h_);
  }

  // adds effect
  void add_effect(const sox_effect_handler_t& eh, int argc, char* const argv[]);

  // adds standard effect identified by name
  void add_effect(const char* name, int argc, char* const argv[]) {
    const sox_effect_handler_t* eh = sox_find_effect(name);
    if (!eh) {
      std::ostringstream os;
      os << "Could not find effect " << name;
      throw std::runtime_error(os.str());
    }
    add_effect(*eh, argc, argv);
  }
};

void chain_handle::add_effect(const sox_effect_handler_t& eh, int argc,
                              char* const argv[]) {
  effect_handle e(sox_create_effect(&eh));
  if (!e.h_) {
    std::ostringstream os;
    os << "Error creating effect " << eh.name;
    throw std::runtime_error(os.str());
  }
  if (argc && sox_effect_options(e.h_, argc, argv) != SOX_SUCCESS) {
    std::ostringstream os;
    os << "Error setting options for effect " << eh.name;
    throw std::runtime_error(os.str());
  }

  // the first effect should receive the original signal description as output
  // description
  if (sox_add_effect(h_, e.h_, &sig_in_, empty_ ? &sig_in_ : sig_out_)
      != SOX_SUCCESS) {
    std::ostringstream os;
    os << "Error adding effect " << eh.name;
    throw std::runtime_error(os.str());
  }
  empty_ = false;

  // once the effect has been added to the chain, it will take care of deleting
  // it when it is deleted itself
  e.h_ = 0;
}


// the state of a chartsweep in progress
class chartsweep {
public:
  // sampling rate of the result file (could be changed through an option)
  static const size_t ext_rate_ = 44100;

  // number of channels in the result file, it cannot be changed that easily
  // because the crossfade logic assumes two channels
  static const size_t ext_channels_ = 2;

private:
  // peak finding parameters
  // start of the range for peak finding, as a factor of a track's length
  const double search_start_;

  // end of the range for peak finding, as a factor of a track's length
  const double search_stop_;

  // hop size for peak finding, in seconds
  const double chunk_dur_;

  // number of hops in a peak finding window
  const size_t chunk_num_;

  // half the window size in seconds
  const double peak_offset_;


  // output parameters
  // the duration of each extract, in seconds
  const double ext_dur_;

  // the duration of the crossfade between extracts, in seconds
  const double fade_dur_;

  // the position of the peak within each extract, in seconds from its start
  const double peak_pos_;

  // the size of an extract, in samples
  const size_t ext_len_;

  // the size of the fade buffer, in samples
  const size_t fade_len_;


  // state variables for peak finding
  // buffer for audio samples during peak finding
  std::vector<sox_sample_t> chunk_;

  // ring buffer for peak finding, holds the energy of the current window, with
  // one value for each hop sized chunk
  double * const chunk_energy_;


  // state variables while writing output
  // buffer holding the fade in and current extract while building the output
  sox_sample_t * const extract_;

  // buffer holding the fade out sections of the current and previous extracts
  sox_sample_t * const fade_buf_;

  // start of the fade out of the previous extract (in fade_buf_)
  // it is set at 0 at first, to do a single fade in
  sox_sample_t *prev_fade_;

  // start of the fade out of the current extract (in fade_buf_)
  sox_sample_t *next_fade_;

  // current write position within extract_
  size_t store_pos_;

  // size of the current extract, only needed because the first one is shorted
  // than ext_len_, to have the total length a multiple of ext_len_
  size_t store_end_;

  // current write position within next_fade_
  size_t fade_pos_;


  // sox effect handlers to store resampled extract in extract_ and next_fade_
  // the only option is the address of the current chartsweep
  static int options(sox_effect_t* effp, int argc, char* argv[]);
  static int start(sox_effect_t* effp);
  static int flow(sox_effect_t* effp, const sox_sample_t* ibuf, sox_sample_t*,
                  size_t* isamp, size_t* osamp);
  static sox_effect_handler_t effect_handler;

  // reads a hop sized chunk of samples into chunk_ buffer during peak finding
  // returns whether a complete chunk could be read
  bool read_chunk(sox_format_t* in);

  // reads a chunk and returns its total energy
  double get_energy(sox_format_t* in);

  // finds peak based on maximum energy, returns its position in seconds within
  // the track
  double find_peak(const char* filename);

  // appends a resampled extract to the output, based on the peak location in
  // seconds
  void append_extract(const char* filename, double peak_time,
                      sox_format_t* out);

public:
  // search_start and search_stop (both as factors of the track length) define
  //   where the peak finding takes place 
  // chunk_dur is the peak finding hop in seconds
  // chunk_num specifies the peak finding window as a number of hops
  // ext_dur is the duration of each extract in seconds
  // fade_dur is the duration of the crossfade between extracts
  // peak_pos is the position of the peak within an extract, as a factor of the
  //   extract's duration
  chartsweep(double search_start, double search_stop, double chunk_dur,
             size_t chunk_num, double ext_dur, double fade_dur,
             double peak_pos);

  ~chartsweep();

  // find the peak within a track, and adds the extract to the chartsweep output
  void process_file(const char* filename, sox_format_t* out) {
    append_extract(filename, find_peak(filename), out);
  }

  // this should be called after the last file, to fade out the last extract
  void fade_out_last(sox_format_t* out);
};

chartsweep::chartsweep(double search_start, double search_stop,
                       double chunk_dur, size_t chunk_num, double ext_dur,
                       double fade_dur, double peak_pos)
  : search_start_(search_start), search_stop_(search_stop),
    chunk_dur_(chunk_dur), chunk_num_(chunk_num),

    // half peak finding window duration
    peak_offset_((chunk_dur_ * chunk_num_) * 0.5),

    ext_dur_(ext_dur), fade_dur_(fade_dur),

    // concert peak position to seconds from start of the extract
    peak_pos_(peak_pos * ext_dur_),

    // convert extract and fade durations to samples
    ext_len_(static_cast<size_t>(ext_dur_ * ext_rate_) * ext_channels_),
    fade_len_(static_cast<size_t>(fade_dur_ * ext_rate_) * ext_channels_),

    // allocate energy ring buffer and extract read buffer
    chunk_energy_(new double[chunk_num_]), extract_(new sox_sample_t[ext_len_]),

    // allocate space for two fade out buffer
    fade_buf_(fade_len_ ? new sox_sample_t[fade_len_ * 2] : 0),

    // there is no previous fade out, set next fade out at second buffer
    prev_fade_(0), next_fade_(fade_len_ ? fade_buf_ + fade_len_ : 0) {
}

chartsweep::~chartsweep() {
  delete [] fade_buf_;
  delete [] extract_;
  delete [] chunk_energy_;
}

// the only option is the chartsweep state
int chartsweep::options(sox_effect_t* effp, int argc, char* argv[]) {
  if (argc != 2 || !argv[1])
    return SOX_EINVAL;
  *reinterpret_cast<chartsweep**>(effp->priv)
    = reinterpret_cast<chartsweep*>(argv[1]);
  return SOX_SUCCESS;
}

int chartsweep::start(sox_effect_t* effp) {
  chartsweep& c = **reinterpret_cast<chartsweep**>(effp->priv);

  // reset write positions
  c.store_pos_ = 0;
  c.fade_pos_ = 0;

  // read fewer samples from the first file to preserve overall length
  // this is probably not all that important, removing it would make the logic
  // simpler (see also computation of duration at the start of append_extract)
  c.store_end_ = c.ext_len_;
  if (!c.prev_fade_ && c.fade_len_)
    c.store_end_ -= c.fade_len_;

  return SOX_SUCCESS;
}

// stores the extract and fade out part
int chartsweep::flow(sox_effect_t* effp, const sox_sample_t* ibuf,
                     sox_sample_t*, size_t* isamp, size_t* osamp) {
  chartsweep& c = **reinterpret_cast<chartsweep**>(effp->priv);
  *osamp = 0;
  size_t to_store, to_fade;

  // store samples in the extract buffer until it is full
  if (c.store_pos_ < c.store_end_) {
    size_t end_pos = c.store_pos_ + *isamp;
    if (end_pos <= c.store_end_) {

      // all available samples will fit in the extract buffer
      to_store = *isamp;
      to_fade = 0;
    }
    else {

      // fill the extract buffer, any additional samples go to the fade buffer
      to_store = c.store_end_ - c.store_pos_;
      to_fade = *isamp - to_store;
    }

    // copy samples to the extract buffer and adjust write position
    std::memcpy(c.extract_ + c.store_pos_, ibuf, to_store * sizeof(sox_sample_t));
    c.store_pos_ += to_store;
  }
  else if (c.fade_pos_ == c.fade_len_) {

    // both extract and fade buffer are full, store nothing
    *isamp = 0;
    return SOX_EOF;
  }
  else {

    // the extract buffer is full, write only in the fade buffer
    to_store = 0;
    to_fade = *isamp;
  }
  size_t end_pos = c.fade_pos_ + to_fade;
  if (end_pos > c.fade_len_) {

    // there is not enough room in the fade buffer, adjust the number of stored
    // samples
    end_pos = c.fade_len_;
    to_fade = end_pos - c.fade_pos_;
  }
  if (to_fade) {

    // write to fade buffer and adjust write position
    std::memcpy(c.next_fade_ + c.fade_pos_, ibuf + to_store,
                to_fade * sizeof(sox_sample_t));
    c.fade_pos_ = end_pos;
  }

  // the total number of samples actually stored in buffers
  *isamp = to_store + to_fade;
  return SOX_SUCCESS;
}

sox_effect_handler_t chartsweep::effect_handler = {
  "chartsweep", NULL, SOX_EFF_MCHAN, options, start, flow, NULL, NULL,
  NULL, sizeof(chartsweep*)
};

void chartsweep::append_extract(const char* filename, double peak_time,
                                sox_format_t* out) {
  std::cout << "peak energy at " << peak_time << " ... " << std::flush;
  sox_format_t* in = sox_open_read(filename, NULL, NULL, NULL);
  if (!in) {
    std::ostringstream os;
    os << "Error opening file \"" << filename << '"';
    throw std::runtime_error(os.str());
  }
  file_handle in_h(in);

  // adjust start time to have peak in defined position after fade in
  double start_time = peak_time - peak_pos_ - fade_dur_;
  if (start_time < 0.)
    start_time = 0.;

  // usually read enough for the extract plus crossfade, except for the first
  // one (to preserve overall length, not such a great idea maybe, see also
  // chartweep::start and the corresponding adjustment of store_end_)
  double duration = ext_dur_;
  if (prev_fade_)
    duration += fade_dur_;

  // make sure it won't extend past end of file
  uint64_t start_pos = start_time * in->signal.rate * in->signal.channels;
  size_t extract_length = in->signal.rate * in->signal.channels * duration;
  if (extract_length >= in->signal.length)
    throw std::runtime_error("file is too short");
  if (start_pos + extract_length >= in->signal.length) {

    // adjust start of fade in so that the extract goes until the end of the
    // file but no further
    start_pos = in->signal.length - extract_length;
    start_time = start_pos / (in->signal.rate * in->signal.channels);
  }

  // try to seek to start position, otherwise trim will skip beginning of file
  if (in->seekable && sox_seek(in, start_pos, SOX_SEEK_SET) == SOX_SUCCESS)
    start_time = 0;

  // set up the sox effect chain that will read and store the extract
  // let sox determine the encoding based on the file name
  sox_encodinginfo_t ext_encoding;
  sox_init_encodinginfo(&ext_encoding);
  ext_encoding.encoding = SOX_DEFAULT_ENCODING;
  ext_encoding.bits_per_sample = 0;
  ext_encoding.compression = 0;
  ext_encoding.opposite_endian = sox_false;

  // but force the final sampling rate and number of channels so that they are
  // consistent for all extracts
  sox_signalinfo_t ext_signal;
  ext_signal.rate = ext_rate_;
  ext_signal.channels = ext_channels_;
  ext_signal.precision = 16;
  ext_signal.length = SOX_IGNORE_LENGTH;
  ext_signal.mult = 0;

  chain_handle chain(&in->encoding, &ext_encoding, &in->signal, &ext_signal);

  // reads input file
  char* effect_args[2];
  effect_args[0] = reinterpret_cast<char*>(in);
  chain.add_effect("input", 1, effect_args);

  // select the extract, if seek failed and start_time is not 0 this will skip
  // the start of the file until start_time (assuming a failed seek didn't move
  // the current position, but the SoX code seems to imply that; if it turns out
  // wrong, the file should be closed and reopened when seek fails)
  std::string start_time_buf;
  std::ostringstream os;
  os << start_time;
  start_time_buf = os.str();
  effect_args[0] = const_cast<char*>(start_time_buf.c_str());

  std::string empty;
  std::string dur_buf;
  os.str(empty);
  os << duration;
  dur_buf = os.str();
  effect_args[1] = const_cast<char*>(dur_buf.c_str());
  chain.add_effect("trim", 2, effect_args);

  // mix channels if there are more than two (to avoid resampling more channels
  // than necessary)
  std::string channels_buf;
  if (in->signal.channels != ext_channels_) {
    os.str(empty);
    channels_buf = os.str();
  }
  if (in->signal.channels > ext_channels_) {
    effect_args[0] = const_cast<char*>(channels_buf.c_str());
    chain.add_effect("channels", 1, effect_args);
  }

  // resample if sampling rate is different from our target rate
  std::string sr_buf;
  if (in->signal.rate != ext_rate_) {
    os.str(empty);
    os << ext_rate_;
    sr_buf = os.str();
    effect_args[0] = const_cast<char*>(sr_buf.c_str());
    chain.add_effect("rate", 1, effect_args);
  }

  // duplicate mono channel (after resampling it, for efficiency's sake)
  if (in->signal.channels < ext_channels_) {
    effect_args[0] = const_cast<char*>(channels_buf.c_str());
    chain.add_effect("channels", 1, effect_args);
  }

  // store result in extract and fade buffers
  effect_args[0] = reinterpret_cast<char*>(this);
  chain.add_effect(effect_handler, 1, effect_args);

  // process everything
  sox_flow_effects(chain.h_, NULL, NULL);
  if (fade_pos_ != fade_len_)
    throw std::runtime_error("error while reading file");

  SOX_SAMPLE_LOCALS;
  size_t clips = 0;
  if (prev_fade_) {
    // there was a previous extract, crossfade its end (which is the previous
    // fade buffer) with the start of the current extract, overwriting the
    // latter
    double pos = 0.;

    // to support a different number of channels, there should be an inner loop
    // and pos should be incremented by the number of channels
    for (sox_sample_t *fade_in = extract_, *fade_end = extract_ + fade_len_,
                      *fade_out = prev_fade_;
         fade_in != fade_end;
         pos += 2) {

      // basic linear crossfade coefficients
      const double ifac = pos / fade_len_;
      const double ofac = 1. - ifac;

      // first channel
      double s = SOX_SAMPLE_TO_FLOAT_64BIT(*fade_in, clips) * ifac;
      s += SOX_SAMPLE_TO_FLOAT_64BIT(*fade_out, clips) * ofac;
      *fade_in = SOX_FLOAT_64BIT_TO_SAMPLE(s, clips);

      // second channel
      s = SOX_SAMPLE_TO_FLOAT_64BIT(*++fade_in, clips) * ifac;
      s += SOX_SAMPLE_TO_FLOAT_64BIT(*++fade_out, clips) * ofac;
      *fade_in = SOX_FLOAT_64BIT_TO_SAMPLE(s, clips);
      ++fade_in;
      ++fade_out;
    }

    // the end of the current extract will be the next fade out, and the end of
    // the previous extract will overwritten by the end of the next one
    std::swap(next_fade_, prev_fade_);
  }
  else if (fade_len_) {
    // this is the first extract, just fade in (in place) at the beginning
    double pos = 0.;

    // to support a different number of channels, there should be an inner loop
    // and pos should be incremented by the number of channels
    for (sox_sample_t *fade_in = extract_, *fade_end = extract_ + fade_len_;
         fade_in != fade_end;
         pos += 2) {

      // basic linear fade in coefficient
      const double fac = pos / fade_len_;

      // first channel
      double s = SOX_SAMPLE_TO_FLOAT_64BIT(*fade_in, clips) * fac;
      *fade_in = SOX_FLOAT_64BIT_TO_SAMPLE(s, clips);

      // second channel
      s = SOX_SAMPLE_TO_FLOAT_64BIT(*++fade_in, clips) * fac;
      *fade_in = SOX_FLOAT_64BIT_TO_SAMPLE(s, clips);
      ++fade_in;
    }

    // the end of the current extract will be the next fade out
    prev_fade_ = next_fade_;

    // setup fade buffer for next extract
    next_fade_ = fade_buf_;
  }

  // write the extract to the output file, including the initial fadein or
  // crossfade, but not the fade out that will be crossfaded with the next
  // extract (using store_end_ because the first extract is shorter)
  if (sox_write(out, extract_, store_end_) != store_end_)
    throw std::runtime_error("error writing to result file");
}

void chartsweep::fade_out_last(sox_format_t* out) {
  if (prev_fade_) {
    // fade out the end of the previous extract
    SOX_SAMPLE_LOCALS;
    size_t clips = 0;
    double pos = 0.;  

    // to support a different number of channels, there should be an inner loop
    // and pos should be incremented by the number of channels
    for (sox_sample_t *fade_out = prev_fade_,
                      *fade_end = prev_fade_ + fade_len_;
         fade_out != fade_end;
         pos += 2) {

      // basic linear fade out
      const double ofac = 1. - (pos / fade_len_);

      // first channel
      double s = SOX_SAMPLE_TO_FLOAT_64BIT(*fade_out, clips) * ofac;
      *fade_out = SOX_FLOAT_64BIT_TO_SAMPLE(s, clips);

      // second channel
      s = SOX_SAMPLE_TO_FLOAT_64BIT(*++fade_out, clips) * ofac;
      *fade_out = SOX_FLOAT_64BIT_TO_SAMPLE(s, clips);
      ++fade_out;
    }

    // write last fade out
    if (sox_write(out, prev_fade_, fade_len_) != fade_len_)
      throw std::runtime_error("error writing to result file");
  }
}

// accumulation functor to compute total energy
struct add_energy : public std::binary_function<double, sox_sample_t, double> {
  double operator()(double acc, sox_sample_t sample) const {
    SOX_SAMPLE_LOCALS;
    size_t clips;
    double amp = SOX_SAMPLE_TO_FLOAT_64BIT(sample, clips);

    // get rid of sox induced warning
    static_cast<void>(clips);

    // use basic signal energy as square of amplitude
    return acc + amp * amp;
  }
};

bool chartsweep::read_chunk(sox_format_t* in) {

  // the number of samples to be read for a full chunk
  size_t left = chunk_.size();

  // current write position in buffer
  sox_sample_t* next_sample = &chunk_[0];

  // loop because sox_read may read fewer samples than requested, without that
  // being an error or meaning the end of file was reached
  do {
    size_t samples = sox_read(in, next_sample, left);

    // end of file means the chunk was not complete
    if (samples == 0)
      return false;

    // adjust write position and number of samples left
    next_sample += samples;
    left -= samples;
  }
  while (left);

  // a full chunk was read
  return true;
}

// attempt to read a chunk and return its total energy as the sum of the squared
// amplitudes
// the standard RMS would be the mean over the channels of the square root of
// the sum of squared amplitudes, but when looking for the position of the
// maximum the mean and square roots do not change the result
// return -1 if a full chunk could not be read
double chartsweep::get_energy(sox_format_t* in) {
  return read_chunk(in)
         ? std::accumulate(chunk_.begin(), chunk_.end(), 0., add_energy())
         : -1.;
}

double chartsweep::find_peak(const char* filename) {
  sox_format_t* in = sox_open_read(filename, NULL, NULL, NULL);
  if (!in) {
    std::ostringstream os;
    os << "Error opening file \"" << filename << '"';
    throw std::runtime_error(os.str());
  }
  file_handle in_h(in);

  // convert chunk duration to a length in frames
  const size_t chunk_len = static_cast<size_t>(in->signal.rate * chunk_dur_);

  // each frame contains one sample per channel
  chunk_.resize(chunk_len * static_cast<size_t>(in->signal.channels));

  // compute search limits in terms of window starting points
  const size_t sig_len = in->signal.length / in->signal.channels;
  const double half_win = chunk_len * chunk_num_ * 0.5;
  long min_start = static_cast<long>(sig_len * search_start_ - half_win);
  if (min_start < 0)
    min_start = 0;
  long max_start = static_cast<long>(sig_len * search_stop_ - half_win);

  // skip until eligible position
  long cur_start;
  for (cur_start = 0; cur_start < min_start; cur_start += chunk_len)
    if (!read_chunk(in))
      throw std::runtime_error("file is too short");

  // filling the ring buffer with the first window's chunks
  double* next_energy = chunk_energy_;
  double* const energy_end = next_energy + chunk_num_;
  for(;next_energy != energy_end; ++next_energy) {
    *next_energy = get_energy(in);
    if (*next_energy < 0.)
      throw std::runtime_error("file is too short");
  }

  // reset next position to oldest value in ring buffer
  next_energy = chunk_energy_;

  // compute first energy value, the initial maximum
  double top_energy = std::accumulate(chunk_energy_, energy_end, 0.);
  long top_start = cur_start;

  // evaluate energy while in eligible range
  for (cur_start += chunk_len;
       cur_start <= max_start && (*next_energy = get_energy(in)) >= 0.;
       cur_start += chunk_len) {

    // energy in window is sum of energy in its chunks
    double energy = std::accumulate(chunk_energy_, energy_end, 0.);
    if (energy > top_energy) {

      // update best energy and position
      top_energy = energy;
      top_start = cur_start;
    }

    // oldest value can be discarded, reset position when at end of ring buffer
    if (++next_energy == energy_end)
      next_energy = chunk_energy_;
  }

  // peak time is at center of the window
  return static_cast<double>(top_start) / in->signal.rate + peak_offset_;
}
}

int main(int argc, const char* argv[]) {

  // default parameter values
  double chunk_dur = 0.1, start_at = 0.1, stop_at = 0.9, ext_dur = 10.,
         fade_dur = 1., peak_pos = 0.1;
  size_t chunk_num = 3;

  // process flags that set those parameter values (this all processing isn't
  // robust, especially number parsing that doesn't check the full string has
  // indeed been consumed)
  int pl_arg = 1;
  while (pl_arg < argc - 2 && argv[pl_arg][0] == '-') {
    switch (argv[pl_arg][1]) {

    // start/end of search and peak position within extract are percentages
    case 'b':
    case 'e':
    case 'p':
      if (argv[pl_arg][2] == 0) {
        double& limit = argv[pl_arg][1] == 'b'
                        ? start_at
                        : (argv[pl_arg][1] == 'e' ? stop_at : peak_pos);
        if (pl_arg < argc - 3) {

          // convert percentage to factor
          limit = atof(argv[pl_arg + 1]) * 0.01;
          if (limit < 0. || limit > 1.) {
            std::cerr << "invalid value " << argv[pl_arg + 1] << " for option "
                      << argv[pl_arg] << " (should be between 0 and 100)"
                      << std::endl;
            return -1;
          }
          pl_arg += 2;
          continue;
        }
        else {
          std::cerr << "missing argument to option " << argv[pl_arg]
                    << std::endl;
          return -1;
        }
      }
      break;

    // fade and extract durations are in seconds
    case 'f':
    case 'l':
      if (argv[pl_arg][2] == 0) {
        if (pl_arg < argc - 3) {
          double& dur = argv[pl_arg][1] == 'l' ? ext_dur : fade_dur;
          dur = atof(argv[pl_arg + 1]);

          // fade can be turned off, but not the extract length
          if (dur < 0. || (dur == 0. && argv[pl_arg][1] == 'l')) {
            std::cerr << "invalid duration " << argv[pl_arg + 1]
                      << " for option " << argv[pl_arg]
                      << " (should be in seconds)"
                      << std::endl;
            return -1;
          }
          pl_arg += 2;
          continue;
        }
        else {
          std::cerr << "missing argument to option " << argv[pl_arg]
                    << std::endl;
          return -1;
        }
      }
      break;

    // step is also in seconds (this could be merged with 'f' and 'l')
    case 's':
      if (argv[pl_arg][2] == 0) {
        if (pl_arg < argc - 3) {
          chunk_dur = atof(argv[++pl_arg]);
          if (chunk_dur <= 0) {
            std::cerr << "invalid step " << argv[pl_arg]
                      << " (should be positive)" << std::endl;
            return -1;
          }
          ++pl_arg;
          continue;
        }
        else {
          std::cerr << "missing argument to option -s" << std::endl;
          return -1;
        }
      }
      break;

    // window size is a number of chunks
    case 'w':
      if (argv[pl_arg][2] == 0) {
        if (pl_arg < argc - 3) {
          int tmp = atoi(argv[++pl_arg]);
          if (tmp <= 0) {
            std::cerr << "invalid window " << argv[pl_arg]
                      << " (should be positive integer)" << std::endl;
            return -1;
          }
          chunk_num = static_cast<size_t>(tmp);
          ++pl_arg;
          continue;
        }
        else {
          std::cerr << "missing argument to option -w" << std::endl;
          return -1;
        }
      }
      break;
    }
    std::cerr << "unknown option " << argv[pl_arg] << std::endl;
    return -1;
  }

  // basic consistency checking
  if (start_at >= stop_at) {
    std::cerr << "inconsistent values for -b and -e options" << std::endl;
    return -1;
  }
  if (fade_dur >= ext_dur) {
    std::cerr << "inconsistent values for -f and -l options" << std::endl;
    return -1;
  }

  // there should be exactly two arguments left, the playlist and result file
  if (pl_arg != argc - 2) {
    std::cerr << "Usage "<<argv[0]<<" [options] playlist out_file"<<std::endl
              << "Version: 1.0" << std::endl
              << "Available options:" << std::endl
              << " -b: % of file where to begin looking for peak (default: 10)"
              << std::endl
              << " -e: % of file where to end looking for peak (default: 90)"
              << std::endl
              << " -f: length of crossfade in seconds (default: 1)"
              << std::endl
              << " -l: length of each extract in seconds (default: 10)"
              << std::endl
              << " -p: peak position in the extract (in % of extract after fade in, default: 10)"
              << std::endl
              << " -s: step (in seconds) for energy function (default: 0.1)"
              << std::endl
              << " -w: window (in number of steps) for energy function (default: 3)"
              << std::endl;
    return 1;
  }

  // now pl_arg is index of the playlist argument
  std::ifstream playlist(argv[pl_arg]);
  if (!playlist.good()) {
    std::cerr << "Could not open playlist \"" << argv[pl_arg] << '"'
              << std::endl;
    return -2;
  }

  if (sox_init() != SOX_SUCCESS) {
    std::cerr << "Error in sox_init " << std::endl;
    return -3;
  }

  // let SoX derive the encoding from the file name, but force the sampling
  // rate, sample precision and numbers of channels to be independent of the
  // input files
  sox_signalinfo_t out_info;
  out_info.rate = chartsweep::ext_rate_;
  out_info.channels = chartsweep::ext_channels_;
  out_info.precision = 16;
  out_info.length = SOX_IGNORE_LENGTH;
  out_info.mult = 0;

  // here it could be nice to refuse to overwrite depending on a flag...
  sox_format_t* out = sox_open_write(argv[pl_arg + 1], &out_info, NULL, NULL,
                                     NULL, NULL);
  if (!out) {
    std::cerr << "Could not open file \"" << argv[pl_arg + 1] << '"'
              << std::endl;
    return -4;
  }

  // retun 0 on success
  int ret = 0;

  // total number of files in the playlist and how many were successfully
  // processed
  int total = 0, ok = 0;
  try {
    file_handle oh(out);
    std::string filename;
    chartsweep sweeper(start_at, stop_at, chunk_dur, chunk_num, ext_dur,
                       fade_dur, peak_pos);
    while (playlist.good()) {
      std::ifstream::int_type c;

      // skip empty lines, accepting both CR and LF as line ends
      do c = playlist.get();
      while (c == '\r' || c =='\n');

      if (c != std::ifstream::traits_type::eof()) {
        if (c == '#') {
          // skip M3U-style comments: ignore the rest of the line
          do c = playlist.get();
          while (c != std::ifstream::traits_type::eof()
                 && c != '\r' && c != '\n');
        }
        else {
          // read file path
          do {
            filename.push_back(static_cast<char>(c));
            c = playlist.get();
          }
          while (c != std::ifstream::traits_type::eof()
                 && c != '\r' && c != '\n');

          // maybe the path should be trimmed here...
        }
      }

      if (!filename.empty()) {
        try {
          ++total;
          std::cout << "processing \"" << filename << "\" ... " << std::flush;

          // try to process the file, any error will throw
          sweeper.process_file(filename.c_str(), out);
          std::cout << "OK" << std::endl;
          ++ok;
        }
        catch (const std::exception& e) {

          // notify the error, the ok counter will not have been incremented
          std::cout << " FAILED: " << e.what() << std::endl;
        }
        filename.clear();
      }
    }

    // fade out the end of the last processed file
    sweeper.fade_out_last(out);
  }
  catch (const std::exception& e) {

    // this is not a file related error, maybe on parsing the playlist...
    std::cerr << e.what() << std::endl;

    // don't exit right away, deinit the library first
    ret = -5;
  }
  
  sox_quit();
  std::cout << std::endl << "Successfully processed " << ok << " files (out of "
            << total << ")" << std::endl << std::endl
            << "Your supercut is complete! Go to " << argv[pl_arg + 1]
            << " and listen" << std::endl;

  return ret;
}

